require "pp"
require "server"
require "observer"
require "serializable"
require "fileutils"
require "conf_file_parser"
require "ext/platform"
require "current_method_name"
require 'asctl/shell'
require "asctl/question_helper"
require "asctl/upgradable"

require 'openssl'
require 'digest'

class RetryError < RuntimeError
end

class Apache < Server
  include Observable
  include Serializable
  include CurrentMethodName
  include Asctl
  include Asctl::QuestionHelper
  include Asctl::Upgradable

  self.file_system_path = __FILE__

  attr_accessor :base_dir
  config_accessor :log_level, "error"
  config_accessor :hostname, nil
  config_accessor :task_status, {}
  config_accessor :ip, ALL_INTERFACES
  config_accessor :includes, []
  config_accessor :ssl_enabled, true
  config_accessor :http_port, 80
  config_accessor :https_port, 443
  config_accessor :admin_email, 'admin@example.com'
  config_accessor :balancer_port, 8080
  config_accessor :httpd_account, ENV['ASPERA_APACHE_USER'] || 'ashttpd'
  config_accessor :httpd_group_account, ENV['ASPERA_APACHE_GROUP'] || 'ashttpd'
  config_accessor :use_ssl_chain_file, false

  # TODO: figure out how to version apache and mysql ourselves
  # (versioning our *use* of them, not the actual apps)
  # config_accessor :version, '0.0.0.1'  # mark us as initially needing upgrade until we figure out the versioning above
  config_accessor :version, '0.0.0.0'

  def initialize
    @base_dir = File.dirname(File.expand_path(__FILE__))
    @reminders = []
    @show_port_reminder = false
    # @executable = File.join(@base_dir,"bin","httpd"+Config::CONFIG["EXEEXT"])
    # @config_files = conf
    # @server_binary = %Q["#{@executable}" -k runservice]
    # @service_dependencies = ["Tcpip","Afd"]
    # @process_id = File.read(File.join(@base_dir,"logs","httpd.pid")) rescue nil
    load_config(default_config)
    super
  end

  def name
    "Apache"
  end

  def server_name
    "Apache HTTPD Server (Aspera)"
  end

  # this should probably be separated into it's own file.
  def make_ssl_cert(hostname)
    options = {
      :cert_digest_algorithm => 'SHA256',
      :cert_extensions => {
        'basicConstraints'      => 'CA:FALSE',
        'subjectKeyIdentifier'  => 'hash',
        'keyUsage'              => 'keyEncipherment,dataEncipherment,digitalSignature',
      },
      :cert_life_years => 10,
      :key_length => 1024,
      :key_path => private_key_path,
    }

    cert_subject = OpenSSL::X509::Name.new([["CN", hostname]])

    from_datetime = Time.now
    datearray = from_datetime.to_a
    datearray[5] = datearray[5] + options[:cert_life_years]
    to_datetime = Time.local(*datearray)

    log "  Generating keypair..."

    private_key = OpenSSL::PKey::RSA.new(options[:key_length])
    public_key = private_key.public_key

    log "  Creating certificate..."

    cert=OpenSSL::X509::Certificate.new

    # populate cert attributes
    cert.version = 2
    cert.serial = Time.now.to_i
    cert.public_key = public_key
    cert.not_before = from_datetime
    cert.not_after = to_datetime
    cert.subject = cert_subject

    # add certificate extensions
    log "  Adding certificate extensions..."
    ef = OpenSSL::X509::ExtensionFactory.new(nil, cert)
    cert.extensions = options[:cert_extensions].map { |k, v| ef.create_extension(k, v) }

    # self signed
    log "  Setting issuer to #{cert_subject.inspect}..."
    cert.issuer = cert_subject
    log "  Self-signing..."
    cert.sign(private_key, OpenSSL::Digest::Digest.new(options[:cert_digest_algorithm]))

    # save private key
    log "  Save private key..."
    File.open(private_key_path, 'wb') { |file| file.write(private_key.to_pem) }
    # save certificate
    log "  Save certificate..."
    File.open(certificate_path, 'wb') { |file| file.write(cert.to_pem) }

    self.use_ssl_chain_file = false
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [hostname]})

    needs_restart
  end

  def install_ssl_cert(cert_file, key_file, chain_file)
    log "  Copying certificate file..."
    cp(cert_file, certificate_path)
    log "  Copying private key file..."
    cp(key_file, private_key_path)
    if chain_file.nil?
      self.use_ssl_chain_file = false
    else
      log "  Copying chain file..."
      cp(chain_file, chain_file_path)
      self.use_ssl_chain_file = true
    end
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [cert_file, key_file, chain_file].compact})
    needs_restart
  end

  def start
    begin
      verify_certificate_files
    rescue => e
      puts
      puts "Unable to start Apache due to missing or corrupted file."
      puts "Restore the file from backup or regenerate with 'asctl apache:make_ssl_cert #{hostname}'"
      raise e
    end
    # OK, now attempt to start
    super
  end

  # Fix for trac ticket 5534
  # Note that these regex constants are intentionally loose - just enough
  # to be reasonably sure we have the right file.
  SSL_CERT_FORMAT = /-----BEGIN CERTIFICATE-----.*-----END CERTIFICATE-----/m  unless defined?(SSL_CERT_FORMAT)
  SSL_KEY_FORMAT  = /-----BEGIN.*KEY-----/m unless defined?(SSL_KEY_FORMAT)
  def verify_certificate_files
    raise "Missing SSL certificate: #{certificate_path}" unless Kernel.test(?e,certificate_path)
    raise "Missing SSL key: #{private_key_path}" unless Kernel.test(?e,private_key_path)
    cert = File.read(certificate_path)
    key  = File.read(private_key_path)
    raise "Invalid SSL certificate format" unless cert =~ SSL_CERT_FORMAT
    raise "Invalid SSL KEY format"         unless key  =~ SSL_KEY_FORMAT
  end

  def default_config
    {
      :server_name => "Apache HTTPD Server (Aspera)",
      :admin_email => "admin@localhost",
      :http_port => 80,
      :https_port => 443,
      :balancer_port => 8080,
      :log_level => "error",
      :httpd_account => system_user,
      :httpd_group_account => system_group,
      :use_ssl_chain_file => false
    }
  end

  def conf_template
    if Platform.name == :windows
      @base_dir+"/conf/httpd_template_windows.conf"
    else
      @base_dir+"/conf/httpd_template_linux.conf"
    end
  end

  def conf
    if Platform.name == :windows
      @base_dir+"/conf/httpd.conf"
    else
      @base_dir+"/conf/ashttpd.conf"
    end
  end

  def includes_str
    return "" if includes.empty?
    return 'Include "' + includes.join("\"\nInclude \"") + '"' + "\n"
  end

  def ssl_include_str
    "Include \"#{base_dir}/conf/extra/httpd-ssl.conf\""
  end

  def generate_config
    log "regenerating apache config files"
    h = {
      /\$\{path\}/ => base_dir,
      /\$\{port\}/ => http_port,
      /\$\{https_port\}/ => https_port,
      /\$\{email\}/ => admin_email,
      /\$\{hostname\}/ => hostname,
      /\$\{balancer_port\}/ => balancer_port,
      /\$\{log_level\}/ => log_level,
      /\$\{includes\}/ => includes_str,
      /\$\{ssl_include\}/ => ssl_include_str,
      /\$\{httpd_account\}/ => httpd_account,
      /\$\{httpd_group_account\}/ => httpd_group_account,
      /\$\{disable_chain_file\}/ => use_ssl_chain_file ? "" : "#" # comment out chain file as necessary
    }
    h[/\$\{log_path\}/] = (Platform.name == :windows) ? "" : @base_dir+"/"

    ConfFileParser::generate_conf_file( conf_template, conf, h )
    if ssl_enabled
      ConfFileParser::generate_conf_file( @base_dir+"/conf/extra/httpd-ssl_template.conf",
                                          @base_dir+"/conf/extra/httpd-ssl.conf", h )
    end
    needs_restart
  end

  def add_aspera_version_headers_to_conf
    conf = "#{File.dirname(File.expand_path(__FILE__))}/conf"
    files = Dir.glob("#{conf}/*_template_*.conf") + Dir.glob("#{conf}/ashttpd.conf")
    files.each do |file|
      file_contents = File.read(file)
      if file_contents !~ /X-Aspera/i
        File.open(file, "a") { |f| f.puts; f.puts "Header set X-IBM-Aspera \"F4.4.2-PL3\"" }
      end
    end
  end

  def has_include?( s )
    @config[:includes].include?(s)
  end

  def log_level=( new_level )
    allowed_levels = ["emerg","alert","crit","error","warn","notice","info","debug"]
    # return if new_level == @config[:log_level] # todo: consider avoiding unnecessary restarts
    raise "Invalid log level '#{new_level}'.  Must be one of '"+allowed_levels.join("' '")+"'" if !allowed_levels.include?(new_level)
    @config[:log_level] = new_level
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [new_level]})
  end

  def disable_logs
    self.log_level=("emerg")
  end

  def enable_logs
    self.log_level=("notice")
  end

  def delete_logs_older_than( days )
    days = days.to_i
    raise ArgumentError,"'days' argument must be at least one day." if days < 1
    deleted = []
    Dir.glob("#{@base_dir}/logs/*_log.*") do |log|
      if log =~ /.*\.(\d+)$/
        if Time.at($1.to_i) < (Time.new - (days*24*60*60))
          rm_f(log)
          deleted << log
        end
      end
    end
    deleted
  end

  def add_include( s )
    # return if has_include?(s) # todo: consider avoiding unnecessary restarts
    @config[:includes] ||= []
    @config[:includes] << s
    @config[:includes].uniq!
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [s]})
  end

  def remove_include( s )
    # return if !has_include?(s) # todo: consider avoiding unnecessary restarts
    @config[:includes].delete( s )
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [s]})
  end

  def http_port=( p )
    p = p.strict_to_i
    # return if p == http_port # todo: consider avoiding unnecessary restarts
    validate_http_port(p)
    @config[:http_port] = p
    @show_port_reminder = true
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [p]})
  end

  def validate_http_port(port,https=https_port,balancer=balancer_port)
    validate_port(port,http_port) do |p|
      if p == https
        raise ArgumentError,"HTTP port must not be the same as the HTTPS port."
      elsif p == balancer
        raise ArgumentError,"HTTP port must not be the same as the balancer port."
      elsif !Port.available?(port)
        raise ArgumentError,"Port not available"
      end
    end
  end

  def validate_https_port(port,http=http_port,balancer=balancer_port)
    validate_port(port,https_port) do |p|
      if p == http
        raise ArgumentError,"HTTPS port must not be the same as the HTTP port."
      elsif p == balancer
        raise ArgumentError,"HTTPS port must not be the same as the balancer port."
      elsif !Port.available?(port)
        raise ArgumentError,"Port not available"
      end
    end
  end

  def validate_balancer_port(port,http=http_port,https=https_port)
    validate_port(port,balancer_port) do |p|
      if p == http
        raise ArgumentError,"Balancer port must not be the same as the HTTP port."
      elsif p == https
        raise ArgumentError,"Balancer port must not be the same as the HTTPS port."
      elsif !Port.available?(port)
        raise ArgumentError,"Port not available"
      end
    end
  end

  def validate_port(port,running_port)
    port = port.strict_to_i
    Port.validate!(port)
    if running? && port == running_port
      return true
    end
    yield(port)
  end

  def port_reminder
    "Apache needs ports #{http_port} and #{https_port} open in firewalls."
  end

  def https_port=( p )
    p = p.strict_to_i
    # return if p == https_port # todo: consider avoiding unnecessary restarts
    validate_https_port(p)
    @config[:https_port] = p
    @show_port_reminder = true
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [p]})
  end

  def admin_email=( e )
    # return if e == admin_email # todo: consider avoiding unnecessary restarts
    @config[:admin_email] = e
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [e]})
  end

  def httpd_account=( e )
    # return if e == httpd_account # todo: consider avoiding unnecessary restarts
    @config[:httpd_account] = e
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [e]})
  end

  def balancer_port=( p )
    p = p.strict_to_i
    # return if p == balancer_port # todo: consider avoiding unnecessary restarts
    validate_balancer_port(p)
    @config[:balancer_port] = p
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [p]})
  end

  # def ssl_enabled=( e )
  #   # TODO: validate 'e' iff we support ssl_enabled
  #   @config[:ssl_enabled] = e
  #   generate_config
  #   changed(); notify_observers(self, {:method => current_method_name, :args => [e]})
  # end

  def conf_dir
    File.join(base_dir, 'conf')
  end

  def certificate_path
    File.join(conf_dir, 'server.crt')
  end

  def certificate_exists?
    File.exists? certificate_path
  end

  def backup_certificate
    cp certificate_path, backup_name(certificate_path)
  end

  def private_key_path
    File.join(conf_dir, 'server.key')
  end

  def private_key_exists?
    File.exists? private_key_path
  end

  def backup_private_key
    cp private_key_path, backup_name(private_key_path)
  end

  def chain_file_path
    File.join(conf_dir, 'server-ca.crt')
  end

  def chain_file_exists?
    File.exists? chain_file_path
  end

  def backup_chain_file
    cp chain_file_path, backup_name(chain_file_path)
  end


  def hostname=( e )
    # return if e == hostname # todo: consider avoiding unnecessary restarts
    @config[:hostname] = e
    generate_config
    changed(); notify_observers(self, {:method => current_method_name, :args => [e]})
  end

  def backup_name(path)
    counter = 1
    while File.exist?(path + ".bak.#{counter}")
      counter += 1
    end
    path + ".bak.#{counter}"
  end

  def reload
    if running?
      stop
      start
    end
  end

  def system_user
    system_user_info['system_user'] || ENV['ASPERA_APACHE_USER'] || 'ashttpd'
  end

  def system_group
    system_user_info['system_group'] || ENV['ASPERA_APACHE_GROUP'] || 'ashttpd'
  end

  def non_default_system_user?
    system_user != 'ashttpd'
  end

  def create_system_user
    if !Platform.windows?
      Platform.add_service_user(system_user, system_group)
    end
  end

  def find_open_http_port
    return http_port if running?
    port = 80
    while !Port.available?(port)
      port = rand(8000)+1024
    end
    port
  end

  def mark_as_enabled
    @config[:enabled] = true
  end

  def set_basedir_permissions
    # Fix #7428 by locking down many more files
    unless windows?
      Dir.chdir(@base_dir) do
        recursive_no_access_by_others = "-R o-rwx"
        symbolic_chmod( recursive_no_access_by_others,
          [
            "../asctl",
            # "../lib",
            ".",
          ] )
        # non-recursive:
        symbolic_chmod( "ugo+rX",
          Dir.glob("htdocs/*") +
          Dir.glob("../lib/*") + [
            "htdocs",
            ".",
            "../lib",
          ] )
        chown_R("daemon", "daemon", "htdocs")
      end
    end
  end

  def create_required_folders
    ["#{base_dir}/logs"].each do |d|
      mkdir_p(d)
      chmod(0777, d)
    end
  end

  def restore_rpmsave_files
    File.remove_extension(rpmsave_files)
  end

  def rpmsave_files
    ["#{@base_dir}/htdocs/index.html.rpmsave"].find_all do |file|
      File.exist?(file)
    end
  end

  def version_file
    File.join(@base_dir,'version.txt')
  end

  def last_installed_version_file
    File.join(@base_dir,'last_version.txt')
  end
  # def installed_version
  #   return @installed_version if @installed_version
  #
  #   out,err,st = Asctl::Shell.ex(Asctl::Shell.escape("#{@base_dir}/bin/ashttpd"),
  #                                '-v')
  #   # puts "out=#{out}"
  #   out =~ /version:\sApache\/(\d.*?)\s\(/
  #   @installed_version = $1
  #   # puts "version: #{@installed_version}."
  #   return @installed_version
  # end
  #
  # def last_installed_version_exists?
  #   false
  # end
  # def last_installed_version
  #   Aspera::Version.new("0.0.0.1")
  # end

  def reminders
    return [] if disabled?
    @reminders << port_reminder if @show_port_reminder
    @reminders.uniq
  end

end

if Platform.name != :windows
  load File.dirname(File.expand_path(__FILE__))+"/linux_apache.rb"
end
