require File.dirname(File.expand_path(__FILE__))+"/apache.rb"
require "asctl/app_controller"
require "controller"
require "server_controller"
require "pp"

class ApacheController < Asctl::AppController
  include ServerController
  include Asctl
  include Asctl::QuestionHelper

  def initialize
    super
    @instance = ObjectFactory.load(Apache)
    @cmd = "apache"
  end

  def_command :create_system_user, {:flags => [:check_upgrade, :check_setup]}
  def_command :backup_certificate, {:flags => [:check_upgrade, :check_setup]}
  def_command :backup_private_key, {:flags => [:check_upgrade, :check_setup]}
  def_command :admin_email, {:flags => default_command_flags}   # for the conf file
  # def_command :ssl_enabled, {:flags => [:check_upgrade, :check_setup, :ask_restart]}   # not used yet
  def_command :balancer_port, {:flags => default_command_flags} # do we still use this?
  def_command :set_basedir_permissions, {:flags => [:check_setup, :check_upgrade, :allow_while_disabled]}

  def_command :generate_config, {
    :help => "When the templates are changed, generate_config will re-generate "+
    "Apache's conf files using the templates.",
    :flags => default_command_flags,
  }

  def_command :http_port, {
    :help => "Display/change the insecure HTTP port the web server listens to.",
    :args => "[port]",
    :flags => default_command_flags,
  }

  def_command :https_port, {
    :help => "Display/change the secure HTTPS port the web server listens to.",
    :args => "[port]",
    :flags => default_command_flags,
  }

  def_command :hostname, {
    :help => "Display/change the hostname or ip address of the server.",
    :args => "[hostname]",
    :flags => default_command_flags,
    :block => lambda do |command_array, *arg_array|
      get_or_set(:hostname,*arg_array)
      value = arg_array.first
      if value
        optionally_make_ssl_cert( value )
      end
    end
  }

  def_command :disable_logs, {
    :help => "Sets the log_level to 'emerg'.",
    :flags => default_command_flags,
  }

  def_command :enable_logs, {
    :help => "Sets the log_level to 'notice'.",
    :flags => default_command_flags,
  }

  def_command :log_level, {
    :help => "Display/change Apache HTTPD log level: crit, error, warn, notice, info, debug.",
    :args => "[level]",
    :flags => default_command_flags,
  }

  if !Platform.windows?
    def_command :reload, {
      :help => "Reload Apache's conf without shutting down.",
      :flags => [:check_upgrade, :check_setup],
    }
  end

  def_command :delete_logs_older_than, {
    :help => "Deletes log files older than the specified number of days.",
    :args => "days",
    :flags => [:check_upgrade, :check_setup],
    :block => lambda do |command_array, number_of_days|
      puts_name_and "Deleting log files..."
      a = @instance.delete_logs_older_than( number_of_days )
      puts_name_and "#{a.size} log file(s) deleted."
    end
  }

  def_command :make_ssl_cert, {
    :help => "Create a self-signed SSL certificate for the specified hostname.",
    :args => "hostname",
    :flags => default_command_flags,
    :call => :_make_ssl_cert,
  }

  def_command :install_ssl_cert, {
    :help => "Install the specified SSL certificate.",
    :args => "cert_file key_file [chain_file]",
    :flags => default_command_flags,
    :call => :_install_ssl_cert,
  }

  def fields_for_info
    super + [
      [:http_port  ,  @instance.http_port],
      [:https_port ,  @instance.https_port],
      [:hostname   ,  @instance.hostname],
      [:version    ,  @instance.version],
    ]
  end

  def_command :rpm_install, {
    # :flags => # inherit them
    :block => lambda { |cmd,*args|
      @instance.install
      @instance.set_basedir_permissions
    }
  }

  def_command :rpm_uninstall, {
    # :flags => # inherit them
    :block => lambda { |cmd,*args|
      @instance.uninstall
    }
  }

  def_command :version, {
    :block => lambda { |cmd,*args|
      @in_version = true
      begin
        super
      ensure
        @in_version = false
      end
    }
  }

  def print_name
    if @in_version
      print name + " (Common): "
    else
      super
    end
  end

  def _install_ssl_cert(cert_file, key_file, chain_file = nil)
    puts_name_and "Backing up certificate, private key, and chain file if they exist"
    @instance.backup_certificate if @instance.certificate_exists?
    @instance.backup_private_key if @instance.private_key_exists?
    @instance.backup_chain_file if @instance.chain_file_exists? && chain_file
    puts_name_and "Installing specified SSL certificate"
    @instance.install_ssl_cert(cert_file, key_file, chain_file)
  end

  def _make_ssl_cert(h)
    puts_name_and "Backing up certificate and private key if they exist"
    @instance.backup_certificate if @instance.certificate_exists?
    @instance.backup_private_key if @instance.private_key_exists?
    puts_name_and "Generating self-signed SSL certificate for hostname #{h}"
    @instance.make_ssl_cert h
  end

  def optionally_make_ssl_cert( e )
    if @instance.certificate_exists? || @instance.private_key_exists?
      reply = ask("Would you like to generate a new SSL certificate? (y/n) " ) {|q| q.validate=/y|n/}
      @instance.make_ssl_cert(e) if reply == 'y'
    end
  end

  def instances
    # Originally this meant a list of models - but as we move questions up to the controller, we
    # need instances to hold a list of controllers (actually, the routines using it won't care,
    # they just use duck typing to access the necessary methods).
    #   old return value =>  [@instance]
    [self]
  end

  def puts_name_and(s)
    puts("#{name}: "+s.to_s)
  end

  def verifiable_items(answers={})
    arr = []
    arr << [:enabled,  'Enabled']
    if answers[:enabled]
      arr << [:hostname,      'Hostname']
      arr << [:ip,            'Bind Address']
      arr << [:http_port,     'HTTP Port']
      arr << [:https_port,    'HTTPS Port']
      if answers[:ssl_cert] =~ /^c/i
        arr << [:ssl_cert_path, 'SSL Cert']
        arr << [:ssl_key_path,  'SSL Key']
      end
    end
    arr
  end

  def default_answers
    { :detailed => false,
      :ip => ALL_INTERFACES,
      :http_port => 80,
      :https_port => 443,
      :ssl_enabled => true,
      :ssl_cert => keys_exist? ? 's' : 'g',
    }
  end

  def ask_questions_for_setup( q, new_answers={} )
    @instance.answers = @instance.answers.merge new_answers
    q.answers = @instance.answers
    detailed_or_streamlined?(q)
    q.shift do
      q.answers[:enabled] = true
      if q.answers[:enabled]
        prompt_ip(q,:if => :detailed)
        prompt_ssl_enabled(q,:if => :detailed)  # not really supported - always set to true
        prompt_hostname(q)
        prompt_http_port(q,:if => :detailed)
        prompt_https_port(q,:if => :detailed) # even if ssl is disabled, apps may need the 'external' https port
        if q.answers[:ssl_enabled]
          prompt_ssl_cert(q,:if => :detailed)
        end
      end
    end
    return q.answers
  end

  def upgrade_tasks( answers = {} )
    tasks = []
    tasks << new_task("Restoring rpmsave files") {
      @instance.restore_rpmsave_files
    } unless @instance.rpmsave_files.empty?

    unless Platform.windows? || @instance.non_default_system_user?
      tasks << new_task("Securing daemon user account") do
        if Platform.has_non_false_shell(@instance.system_user)
          Platform.change_shell(@instance.system_user,'/bin/false')
        end
      end
    end

    tasks
  end

  def tasks_for_setup( answers = {} )
    # Pre-Answer Checks ==========================================================
    common_pre_answer_checks

    tasks = []

    tasks << Asctl::TaskTitle.new(self.name)

    tasks << new_task("Recording system user and group") {
      @instance.record_system_user
    }

    tasks += upgrade_tasks(answers) if @instance.upgrading?

    tasks << new_task("Pre-execution checks") do
      if answers[:http_port] && (answers[:http_port] == answers[:https_port])
        raise "The http port can not equal the https port"
      end
      #   re-check:
      if answers[:ssl_cert] =~ /c/
        if answers[:ssl_cert_path]
          raise "SSL Certificate does not exist" if !File.exists?(answers[:ssl_cert_path])
        end
        if answers[:ssl_key_path]
          raise "SSL Key does not exist" if !File.exists?(answers[:ssl_key_path])
        end
      end
    end

    tasks << new_task("Enabling Apache") {
      @instance.enabled = true
    }
    if Platform.windows?
      tasks << new_task("Service starts automatically at boot") do
        @instance.start_automatically
      end
    else
      tasks << new_task("Installing startup script") {
        @instance.install_service if !@instance.service?
      }
      tasks << new_task("Creating system user") {
        @instance.create_system_user
      }
    end
    tasks << new_task("Creating required directories") do
      @instance.create_required_folders
    end
    tasks << new_task("Setting IP") {
      @instance.ip = answers[:ip]
    }
    tasks << new_task("Setting hostname") {
      @instance.hostname = answers[:hostname]
    }
    # We believe that code in Faspex expects SSL to be enabled, so we don't allow it to be disabled.
    # tasks << new_task("Enabling/Disabling SSL") {
    #   @instance.ssl_enabled = answers[:ssl_enabled]
    # }
    tasks << new_task("Setting HTTP port") {
      @instance.http_port = answers[:http_port]
    }
    tasks << new_task("Setting HTTPS port") {
      @instance.https_port = answers[:https_port]
    }
    tasks << new_task("Inserting Aspera Version Headers") do
      @instance.add_aspera_version_headers_to_conf
    end

    # ssl_cert
    if answers[:ssl_enabled] && answers[:ssl_cert] =~ /g|c/
      if answers[:ssl_cert] =~ /c/
        tasks << new_task("Copying SSL cert") do
          cp( answers[:ssl_cert_path], @instance.certificate_path )
        end
        tasks << new_task("Copying SSL key") do
          cp( answers[:ssl_key_path], @instance.private_key_path )
        end
      else
        tasks << new_task("Generating SSL key") { @instance.make_ssl_cert(@instance.hostname) }
      end
    end

    tasks << new_task("Generating configuration file") {
      @instance.generate_config
    }
    if !Platform.windows?
      tasks << new_task("Checking #{@instance.system_user} user") do
        `id #{@instance.system_user} 2>&1`
        raise "#{@instance.system_user} user does not exist." if $? != 0
      end
    end

    tasks << new_task("Updating version") {
      @instance.update_version
    }

    tasks
  end

  # ============================================================================

  def common_pre_answer_checks

    # must be root
    if Platform.name != :windows
      raise "You must be running as root" unless running_as_root?
    else
      raise "You must be an administrator" if false # jwe todo
    end

    # umask
    if Platform.name != :windows
      cmd = 'echo `umask`' # umask is a shell command (like 'export'), not an executable script
      umask,err,status = Shell.ex(cmd)
      raise "You must have umask set to 0022" if umask !~ /^0022/
    end

    #if enabled
    #  # jwe todo: upgrade check
    #  # must run upgrade first
    #  raise "You must run the 'upgrade' command first." if current_version < installed_version
    #end
  end

  def prompt_ip(questionnaire,options = {})
    options[:error_message] ||= "That IP address isn't an address of this computer"
    questionnaire.ask(:ip,"Enter IP address of network interface for Apache to listen on: %s", options) do |str|
      listen_on_all_addresses?(str) || (IPAddr.valid?(str) && interface_exists?(str))
    end
    if listen_on_all_addresses?(questionnaire.answers[:ip])
      questionnaire.answers[:ip] = ALL_INTERFACES
    end
  end

  def prompt_http_port(questionnaire,options = {})
    options[:type] ||= :integer
    questionnaire.ask(:http_port,"What port would you like to run Apache http on? %s", options) do |str|
      begin
        @instance.validate_http_port(str,
          questionnaire.answers[:https_port])
        true
      rescue ArgumentError
        [false,$!.message]
      end
    end
  end

  def prompt_https_port(questionnaire,options = {})
    options[:type] ||= :integer
    http_port_answer = questionnaire.answers[:http_port]
    if questionnaire.answers[:ssl_enabled]
      ssl = true
      question = "What port would you like to run Apache https on? %s"
    else
      ssl = false
      question = "What port will be used for https connections? %s"
    end
    questionnaire.ask(:https_port, question, options) do |str|
      begin
        @instance.validate_https_port(str,
          questionnaire.answers[:http_port])
        if str.to_i == http_port_answer
          raise ArgumentError,"HTTPS port cannot match the HTTP port"
        end
        true
      rescue ArgumentError
        [false,$!.message]
      end
    end
  end

  def keys_exist?
    @instance.private_key_exists? && @instance.certificate_exists?
  end

  # TODO: Support modified customer's apache ssl conf file to point to keys elsewhere in existing installations.
  # TODO: We need to update docs to say "replace generated certs at these paths with your own".
  def prompt_ssl_cert(questionnaire,options = {})
    questionnaire.easy_answers[:ssl_cert] = (keys_exist? ? 's' : 'g')
    begin
      possible_answers = '[g]enerate,[c]opy'
      possible_answers << ',[s]kip' if keys_exist?

      questionnaire.ask(:ssl_cert,
                        "Would you like to generate a self-signed SSL certificate, or install your own "+
                        "(#{possible_answers})%s?", :if => :detailed ) do |str|
        ret = (str =~ /^(g|c|s)$/i)
        if ret && str =~ /s/i
          ret = [false, "Private key does not exist in #{private_key_path}."] if !@instance.private_key_exists?
          ret = [false, "Certificate does not exist in #{certificate_path}."] if !@instance.certificate_exists?
        end
        ret
      end
      if questionnaire.answers[:ssl_cert] =~ /c/
        questionnaire.indent = 4
        prompt_ssl_cert_copy_paths(questionnaire,options)
      elsif questionnaire.answers[:ssl_cert] =~ /g/
        puts "    Key and certificate will be generated in this directory:"
        puts "      #{@instance.conf_dir}"
      end
    rescue RetryError
      questionnaire.indent = 2
      retry
    ensure
      questionnaire.indent = 2
    end
  end

  def check_path(str)
    ret = false
    msg = ""
    if str == ""
      msg = "Path must not be blank."
    elsif !File.exist?(str)
      msg = "File '#{str}' does not exist."
    elsif File.directory?(str)
      msg = "'#{str}' is a directory."
    else
      ret = true
    end
    return [ret, msg]
  end

  # Copying a certificate requires that they tell us what to copy.
  def prompt_ssl_cert_copy_paths(questionnaire,options = {})
    questionnaire.answers[:ssl_cert_path] ||= @instance.certificate_path
    questionnaire.answers[:ssl_key_path]  ||= @instance.private_key_path
    questionnaire.ask(:ssl_cert_path, "Enter the path to the SSL certificate (.crt file) ('b' to go back): %s") do |str|
      raise RetryError, "Retry requested" if str == 'b'
      check_path(str)
    end
    questionnaire.ask(:ssl_key_path, "Enter the path to the private SSL key (.key file) ('b' to go back): %s") do |str|
      raise RetryError, "Retry requested" if str == 'b'
      check_path(str)
    end
  end

  def prompt_hostname(questionnaire,options={})
    if questionnaire.answers[:ssl_enabled]
      question = "What hostname or IP address should Apache use to identify itself (in the SSL Certificate)? %s"
    else
      question = "What is the external hostname or IP address? %s"
    end
    ask_hostname_without_local_address_validation(questionnaire, question, options)
  end

  def ask_hostname_without_local_address_validation(questionnaire,question,options={})
    questionnaire.ask(:hostname, question,options) do |str|
      !str.blank?
    end
    if listen_on_all_addresses?(questionnaire.answers[:ip])
      questionnaire.answers[:hostname] = ALL_INTERFACES
    end
  end


  def prompt_ssl_enabled(questionnaire,options={})
    # # questionnaire.casts(:inverted_yes_no,
    # #                     lambda { |type| type ? 'n' : 'y' },
    # #                     lambda { |str| str =~ /^y/i ? false : true })

    # questionnaire.ask(:ssl_enabled,
    #                   "Enable SSL on this server (y/n)? %s",
    #                   :type => :yes_no)
    questionnaire.answers[:ssl_enabled] = true
  end


end
