class OauthController < SamlController
    EXPIRATION_WINDOW = 1.day
    EXPIRATION_AGO = 1.day.ago
  
    skip_before_filter :check_saml_enabled
  
    def create
      expires_now
      case params[:grant_type]
        when 'password'
          authenticate_basic_auth
        when 'urn:ietf:params:oauth:grant-type:saml2-bearer'
          set_saml_config
          super
        when 'refresh_token'
          handle_refresh_token
        else
          render_oauth_error(:unsupported_grant_type)
      end
    end
  
    private
  
    def authenticate_basic_auth
      if user = basic_auth_user
        if (user.password_expired?)
            render_oauth_error(:expired_password)
        else
            user.update_login_info(request.host, request.remote_ip)
            render_token(create_token_for_user(user))
        end
      else
        render_oauth_error(:invalid_credentials)
      end
    end
  
    def saml_response
      @saml_response ||= OneLogin::RubySaml::Response.new(params[:assertion])
    end
  
    def success
      user = AuthorizableUser.authenticate_with_saml(saml_response, request.remote_ip,email,lookup_id)
      if user.errors.any?
        Rails.logger.warn(user.errors)
        error_message = user.errors.first[1] rescue user.errors
  
        render_oauth_error(:invalid_client, error_message)
      else
        user.update_login_info(request.host, request.remote_ip)
        render_token(create_token_for_user(user))
      end
    end
  
    def failure
      if saml_response.errors.any?{|e| e.include?('The status code of the Response was not Success, was Requester')}
        Rails.logger.error "Invalid SAML Configuration - Might be mismatch between specified name-id format and SAML server name-id format"
        invalid_saml_response
        return
      end
  
      render_oauth_error(:invalid_saml_response, saml_response.status_message)
    end
  
    def handle_refresh_token
      if token = OauthToken.find(:all, :conditions => ["created_at > ? and refresh_token = ?", EXPIRATION_AGO, params[:refresh_token].to_s]).first
        token.delete
        render_token(create_token_for_user(token.user))
      else
        render_oauth_error(:invalid_refresh_token)
      end
    end
  
    def invalid_saml_response
      Rails.logger.warn("SAML assertion was invalid: #{saml_response.errors}")
      render_oauth_error(:invalid_saml_response)
    end
  
    def create_token_for_user(user)
      OauthToken.create(:user => user)
    end
  
    def render_token(token)
      render json: {:access_token  => token.access_token,
                    :token_type    => :bearer,
                    :refresh_token => token.refresh_token,
                    :expires_in    => EXPIRATION_WINDOW}
    end
  
    def render_oauth_error(error, message = nil)
      message ||= error.to_s.humanize
      case error
        when :invalid_client
          render json: {:error => :invalid_client, :error_description => message}, status: 400
        when :invalid_credentials
          render json: {:error => :invalid_credentials, :error_description => message}, status: 401
        when :invalid_saml_response, :invalid_refresh_token
          render json: {:error => :invalid_grant, :error_description => message}, status: 400
        when :unsupported_grant_type
          render json: {:error => :unsupported_grant_type}, status: 400
        when :expired_password
          render json: {:error => :expired_password, :error_description => message + ". Log into the UI to reset it."}, status: 401
      end
    end
  end