class EMembership < ActiveRecord::Base
  include MultiServerHooks
  include MembershipMethods

  belongs_to :user
  belongs_to :e_workgroup
  belongs_to :server_invited_from, :class_name => 'HomeServer'

  named_scope :wg_admins, :conditions => {:is_wg_admin => true}
  named_scope :search_for, (lambda do |search|
    {
      include: :user,
      order: 'name ASC',
      conditions: ['users.name LIKE :search', { search: "%#{search}%" } ]
    }
  end)

  attr_accessible :access_level, :type_individually_set, :e_workgroup, :can_read,
      :dbox_extusr_single_use, :dbox_extusr_expires_at, :dbox_extusr_passcode,
      :dbox_extusr_expired, :dbox_description, :invited_at, :server_invited_from, :user_id,
      :e_workgroup_id, :public, :can_write, :is_wg_admin, :dbox_ear, :prevent_http_upload

  alias_attribute :ear, :dbox_ear
  before_validation_on_create :default_values, if: :enforce_maximum_invitation_duration?

  def self.dbox_ear(passcode)
    passcode = passcode.to_s
    return EMembership.all.select{|element| element.dbox_extusr_passcode == passcode}[0].dbox_ear rescue '0'
  end

  def self.expire_dropboxes
    to_expire = EMembership.all(:conditions => ['dbox_extusr_expires_at < ? AND dbox_extusr_expired = ?', Time.now, false])
    to_expire.each { |m| m.expire! }
  end

  def expired?
    dbox_extusr_expired? || dbox_extusr_expires_at && dbox_extusr_expires_at < Time.now
  end

  def expire!
    self.dbox_extusr_expired = true
    save
  end

  def status
    if user.external_email_user? && expired?
      I18n.t('app.models.e_membership.status.expired')
    elsif user.active?
      I18n.t('app.models.e_membership.status.active')
    else
      I18n.t('app.models.e_membership.status.inactive')
    end
  end

  def sync_attributes
    { :workgroup_name => workgroup_name,
      :user_name => user_name,
      :is_wg_admin => is_wg_admin }
  end

  def user_name
    member = self.user
    if member.nil?
      "#ERROR# - non-existent user (id=" + self.user_id.to_s + ")"
    else
      member.name
    end
  end

  def expiration_desc
    if dbox_extusr_single_use? && !dbox_extusr_expires_at
      "After one upload"
    elsif dbox_extusr_single_use? && dbox_extusr_expires_at
      "After one upload or " + Utils.admin_time_to_human_one_line(dbox_extusr_expires_at) + " " + Time.now.zone
    elsif dbox_extusr_expires_at
      Utils.admin_time_to_human_one_line(dbox_extusr_expires_at) + " " + Time.now.zone
    else
      ""
    end
  end

  # match the behavior of Invitation
  def default_values
    if EConfiguration && EConfiguration.invitation_link_expiration_policy? && EConfiguration.invitation_link_expires? && dbox_extusr_expires_at.blank?
      self.dbox_extusr_expires_at =  Date.today + EConfiguration.invitation_link_days_valid.days
    end
  end

  def enforce_maximum_invitation_duration?
    EConfiguration.enforce_maximum_invitation_duration?
  end

  # match the behavior of Invitation
  def expiration
    if dbox_extusr_single_use? && !dbox_extusr_expires_at
      'after_one_upload'
    elsif dbox_extusr_single_use? && dbox_extusr_expires_at
      'after_one_upload_or_date'
    elsif dbox_extusr_expires_at
      'on_date'
    else
      ''
    end
  end

  # match the behavior of Invitation
  alias_attribute :passcode, :dbox_extusr_passcode
  alias_attribute :expires_at, :dbox_extusr_expires_at
  alias_attribute :description, :dbox_description

  def multi_server_record_identifier
    ""
  end

  def original_multi_server_record_identifier
    ""
  end

  def resend_invitation
    return unless e_workgroup.dropbox? && user.external_email_user?
    update_attributes(:invited_at => Time.now)
    # actual sending will be done by EMembershipObserver on the user's home server when invited_at changes
  end

  def home_server
    home_server_id = user.home_server_id
    if home_server_id == 0 && user.external_email_user? # 'self'
      server_invited_from || HomeServer.default_server # fall back to default server if inviting server is unknown
    else
      user.home_server
    end
  end

  def to_hash
    {
        id: id,
        e_workgroup_id: e_workgroup_id,
        e_workgroup_type: e_workgroup.dropbox ? "Dropbox" : "Workgroup",
        e_workgroup_name: e_workgroup.name,
        user_id: user_id,
        can_read: can_read,
        can_write: can_write,
        is_wg_admin: is_wg_admin,
        created_at: created_at,
        updated_at: updated_at
    }
  end
end
