class Invitation < ActiveRecord::Base
  include MultiServerHooks

  before_validation :sanitize_and_strip_tags
  before_validation :format_expiration
  validates_presence_of :expiration
  validates_date :expires_at, after: Date.today, if: Proc.new{ |i| i.expiration == 'on_date' || i.expiration == 'after_one_upload_or_date' }
  validate :validate_email_address
  validate :validate_expiration_policy
  validates_format_of :email_address, with: Utils::VALID_EMAIL_REGEX, message: "has an invalid format"
  validates_length_of :description, :maximum => 8000, :allow_nil => true
  before_validation_on_create :set_invitee
  before_validation_on_create :default_values, if: :enforce_maximum_invitation_duration?

  attr_accessor :db_id # Dropbox ID

  attr_accessible :email_address,
      :ear,
      :expiration,
      :expires_at,
      :inviter,
      :expired,
      :invited_at,
      :public,
      :description,
      :prevent_http_upload

  belongs_to :inviter, polymorphic: true
  belongs_to :invitee, class_name: "User"
  belongs_to :delivery

  after_save :send_invitation_email, :unless => :expired?

  named_scope :date_expirable, conditions: ['expiration = ? OR expiration = ?', 'on_date', 'after_one_upload_or_date']
  named_scope :private, conditions: ['public = ?', false]
  named_scope :public,  conditions: ['public = ?', true]

  default_value_for :passcode do
    Digest::SHA1.hexdigest(SecureRandom.uuid)
  end

  default_value_for :expiration, 'never'
  default_value_for :description, ""

  def validate_email_address
    if email_address.blank?
      errors.add :email_address, "can't be blank"
    elsif inviter.is_a?(EWorkgroup) && AuthorizableUser.exists?(name: email_address.strip)
      # dropbox invitations cannot be for full-fledged Faspex users
      errors.add_to_base("Cannot invite existing users.")
    end
  end

  def validate_expiration_policy
    db = EWorkgroup.find_by_id(db_id.to_i)
    policy = (db.present? && db.custom_link_expiration_policy?) ? db : EConfiguration

    if policy == EConfiguration && EConfiguration.invitation_link_expiration_policy?
      if !db && !policy.invitation_link_expiration_editable? && !choices_match_policy
        errors.add_to_base("Cannot change default link expiration policy")
      end

      if policy.invitation_link_expires? && expires_at.present? && enforce_maximum_invitation_duration? && Date.today + policy.invitation_link_days_valid.days < expires_at
        errors.add_to_base("Cannot set expiration date more than " + policy.invitation_link_days_valid.to_s + " days in the future")
      end

      if !db && !policy.invitation_link_expiration_editable? && policy.invitation_link_expires?
        self.expires_at = Date.today + policy.invitation_link_days_valid.days
      end
    elsif db.present? && db.custom_link_expiration_policy?
      if policy.invitation_link_expires? && expires_at.present? && Date.today + policy.invitation_link_days_valid.days < expires_at
        errors.add_to_base("Cannot set expiration date more than " + policy.invitation_link_days_valid.to_s + " days in the future")
      end
      if enforce_maximum_invitation_duration? && !(policy.invitation_link_expires? || policy.invitation_link_expires_after_upload?) && expires_at.present? &&
          Date.today + policy.invitation_link_days_valid.days < expires_at
        errors.add_to_base("Cannot set expiration date more than " + policy.invitation_link_days_valid.to_s + " days in the future")
      end
    end
  end

  def choices_match_policy
    case expiration
      when "after_one_upload_or_date"
        EConfiguration.invitation_link_expires? && EConfiguration.invitation_link_expires_after_upload?
      when "after_one_upload"
        !EConfiguration.invitation_link_expires? && EConfiguration.invitation_link_expires_after_upload?
      when "on_date"
        EConfiguration.invitation_link_expires? && !EConfiguration.invitation_link_expires_after_upload?
      when "never"
        !EConfiguration.invitation_link_expires? && !EConfiguration.invitation_link_expires_after_upload?
    end
  end

  def default_values
    db = EWorkgroup.find_by_id(db_id.to_i)
    policy = (db.present? && db.custom_link_expiration_policy?) ? db : EConfiguration
    if EConfiguration && EConfiguration.invitation_link_expiration_policy? && EConfiguration.invitation_link_expires? && expires_at.blank?
      self.expires_at = Date.today + policy.invitation_link_days_valid.days
      if expiration == 'after_one_upload'
        self.expiration = 'after_one_upload_or_date'
      elsif expiration == 'never'
        self.expiration = 'on_date'
      end
    end
  end

  def enforce_maximum_invitation_duration?
    EConfiguration.enforce_maximum_invitation_duration?
  end

  def send_invitation_email
    if invited_at.to_i > invited_at_was.to_i && invitee.home_server == HomeServer.this_server
      case inviter
      when User
        # home_server_id == 0 means any home server
        # if external user is not assigned a specific home server, only send email from inviter's home server
        unless invitee.is_a?(ExternalUser) && invitee.home_server_id == 0 && inviter.home_server != HomeServer.this_server
          BackgroundNotifier.deliver_personal_invitation self
        end
      when EWorkgroup
        # if external user is not assigned a specific home server, only send email from the public URL home server
        unless invitee.is_a?(ExternalUser) && invitee.home_server_id == 0 && EConfiguration.first.public_url_home_server != HomeServer.this_server
          BackgroundNotifier.deliver_dropbox_invitation self
        end
      end
    end
  end

  def multi_server_record_identifier
    ""
  end

  def original_multi_server_record_identifier
    ""
  end

  def single_use?
    expiration == 'after_one_upload' || expiration == 'after_one_upload_or_date'
  end

  def self.expire_invitations
    # some weirdness with comparing dates in mysql, #16121
    #to_expire = Invitation.date_expirable(conditions: ['expires_at <= ? AND expired = ?', Date.today, false])
    #to_expire.each { |i| i.expire! }
    Invitation.date_expirable.each do |i|
      if !i.expired? && i.expires_at <= Date.today
        i.expire!
      end
    end
  end

  def expire!
    self.expired = true
    if HomeServer.this_server.slave?
      self.needs_sync_to_master = true
    end
    save
  end

  alias_attribute :user, :invitee

  def e_workgroup
    if inviter.is_a?(EWorkgroup)
      inviter
    else
      nil
    end
  end

  ##
  # Returns the invitation for the passcode. If maybe_passcode is nil,
  # or the invitation is not found (maybe_passcode is invalid), returns nil.
  def self.invitation_from_passcode(maybe_passcode)
    if maybe_passcode
      #
      # Look up invitation (may be nil)
      #
      return Invitation.find_by_passcode(maybe_passcode.to_s)
    else
      return nil
    end
  end
private

  def set_invitee
    if invitee.nil? && email_address
      self.invitee = User.find_by_name(email_address) ||
          ExternalUser.new(name: email_address, notify_on_download: false, creator_name: inviter.try(:name))
    end
  end

  def format_expiration
    if (self.expiration.include?("after_one_upload") && self.expiration.include?("on_date")) || self.expiration.include?("after_one_upload_or_date")
      self.expiration = "after_one_upload_or_date"
    elsif self.expiration.include?("after_one_upload")
      self.expiration = "after_one_upload"
      self.expires_at = nil
    elsif self.expiration.include?("on_date")
      self.expiration = "on_date"
    else
      self.expiration = "never"
      self.expires_at = nil
    end
  end

  def sanitize_and_strip_tags
    self.description = Utils.sanitize_and_strip_tags(self.description) if self.description_changed?
  end

end
