##
# Some simple methods for parsing request parameters, like the
# user id, package id, and so on.
#
# We put this in a module and mix it in with controllers that need it,
# in order to stay DRY.
#
# Yes, Rails will parse the user id and so on to an integer if it is
# one. But the requester might do /users/foo or something like that.
module Api::RequestParamParsing

  def parse_user
    begin
      #
      # If user_id is 'self', then use current user.
      #
      if params[:user_id] == 'me'
        @user = current_user
        return
      else
        #
        # Otherwise, parse as id
        #
        user_id = params[:user_id].to_i
      end
    rescue
      raise Api::BadRequestError,
            "Bad user id <#{params[:user_id]}> in request."
    end
    @user = User.find_by_id(user_id)
    unless @user
      raise Api::NoContentError.new(
            "User invalid: (user id was #{user_id}).",
                'User not found')
    end
    unless @user == current_user || current_user_is_server_admin
      raise Api::UnauthorizedError.new(
                'You cannot access another user. Only admin can do that.',
                'You cannot access another user. Only admin can do that.')
    end
  end

  def parse_delivery
    begin
      if params[:passcode].nil?
        package_id = params[:package_id].to_i
      else
        recipient_delivery = RecipientDelivery.find_by_send_to_email_passcode(params[:passcode].to_s)
        package_id = recipient_delivery.delivery_id
      end
    rescue
      raise Api::BadRequestError,
            "Bad package id <#{params[:package_id]}> in request."
    end
    @delivery = Delivery.find_and_check_perms(package_id, @user)
    unless @delivery
      raise Api::ContentNotFoundError,
            "Package with id #{package_id} for user #{@user.name} not found."
    end
  end

  def parse_deliveries
    begin
      package_ids = params[:package_ids].map(&:to_i)
    rescue
      raise Api::BadRequestError,
            "Bad package ids <#{params[:package_id]}> in request."
    end
    bad_ids = []
    @deliveries = package_ids.map{ |id|
      d = Delivery.find_and_check_perms(id, @user)
      bad_ids << id unless d.present?
      d
    }.compact

    bad_ids
  end

end
